@@ -49,6 +49,8 @@ enum
   PROP_SHOW_GENERIC_NAMES,
   PROP_SHOW_MENU_ICONS,
   PROP_SHOW_TOOLTIPS,
+  PROP_SHOW_DESKTOP_ACTIONS,
+  PROP_RIGHT_CLICK_EDITS,
   N_PROPERTIES
 };
 
@@ -81,6 +83,8 @@ struct _GarconGtkMenuPrivate
   guint show_generic_names : 1;
   guint show_menu_icons : 1;
   guint show_tooltips : 1;
+  guint show_desktop_actions : 1;
+  guint right_click_edits : 1;
 };
 
 
@@ -167,6 +171,32 @@ garcon_gtk_menu_class_init (GarconGtkMenuClass *klass)
                           G_PARAM_READWRITE
                           | G_PARAM_STATIC_STRINGS);
 
+  /**
+   * GarconMenu:show-desktop-actions:
+   *
+   *
+   **/
+  menu_props[PROP_SHOW_DESKTOP_ACTIONS] =
+    g_param_spec_boolean ("show-desktop-actions",
+                          "show-desktop-actions",
+                          "show desktop actions in a submenu",
+                          FALSE,
+                          G_PARAM_READWRITE
+                          | G_PARAM_STATIC_STRINGS);
+
+  /**
+   * GarconMenu:right-click-edits:
+   *
+   *
+   **/
+  menu_props[PROP_RIGHT_CLICK_EDITS] =
+    g_param_spec_boolean ("right-click-edits",
+                          "right-click-edits",
+                          "right click to edit menu items",
+                          FALSE,
+                          G_PARAM_READWRITE
+                          | G_PARAM_STATIC_STRINGS);
+
   /* install all properties */
   g_object_class_install_properties (gobject_class, N_PROPERTIES, menu_props);
 }
@@ -181,6 +211,8 @@ garcon_gtk_menu_init (GarconGtkMenu *menu)
   menu->priv->show_generic_names = FALSE;
   menu->priv->show_menu_icons = TRUE;
   menu->priv->show_tooltips = FALSE;
+  menu->priv->show_desktop_actions = FALSE;
+  menu->priv->right_click_edits = FALSE;
 
   gtk_menu_set_reserve_toggle_size (GTK_MENU (menu), FALSE);
 }
@@ -231,6 +263,14 @@ garcon_gtk_menu_get_property (GObject    *object,
       g_value_set_boolean (value, menu->priv->show_tooltips);
       break;
 
+    case PROP_SHOW_DESKTOP_ACTIONS:
+      g_value_set_boolean (value, menu->priv->show_desktop_actions);
+      break;
+
+    case PROP_RIGHT_CLICK_EDITS:
+      g_value_set_boolean (value, menu->priv->right_click_edits);
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -265,6 +305,14 @@ garcon_gtk_menu_set_property (GObject      *object,
       garcon_gtk_menu_set_show_tooltips (menu, g_value_get_boolean (value));
       break;
 
+    case PROP_SHOW_DESKTOP_ACTIONS:
+      garcon_gtk_menu_set_show_desktop_actions (menu, g_value_get_boolean (value));
+      break;
+
+    case PROP_RIGHT_CLICK_EDITS:
+      garcon_gtk_menu_set_right_click_edits (menu, g_value_get_boolean (value));
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -301,8 +349,9 @@ garcon_gtk_menu_append_quoted (GString     *string,
 
 
 static void
-garcon_gtk_menu_item_activate (GtkWidget      *mi,
-                               GarconMenuItem *item)
+garcon_gtk_menu_item_activate_real (GtkWidget            *mi,
+                                    GarconMenuItem       *item,
+                                    GarconMenuItemAction *action)
 {
   GString      *string;
   const gchar  *command;
@@ -316,7 +365,15 @@ garcon_gtk_menu_item_activate (GtkWidget      *mi,
   g_return_if_fail (GTK_IS_WIDGET (mi));
   g_return_if_fail (GARCON_IS_MENU_ITEM (item));
 
-  command = garcon_menu_item_get_command (item);
+  if (action != NULL)
+    {
+      command = garcon_menu_item_action_get_command (action);
+    }
+  else
+    {
+      command = garcon_menu_item_get_command (item);
+    }
+
   if (STR_IS_EMPTY (command))
     return;
 
@@ -395,6 +452,95 @@ garcon_gtk_menu_item_activate (GtkWidget      *mi,
 
 
 
+static void
+garcon_gtk_menu_item_edit_launcher (GarconMenuItem *item)
+{
+  GFile   *file;
+  gchar   *uri, *cmd;
+  GError  *error = NULL;
+
+  file = garcon_menu_item_get_file (item);
+
+  if (file)
+    {
+      uri = g_file_get_uri (file);
+      cmd = g_strdup_printf ("exo-desktop-item-edit \"%s\"", uri);
+
+      if (!xfce_spawn_command_line_on_screen (NULL, cmd, FALSE, FALSE, &error))
+        {
+          xfce_message_dialog (NULL,
+                               _("Launch Error"),
+                               "dialog-error",
+                              _("Unable to launch \"exo-desktop-item-edit\", which is required to create and edit menu items."),
+                              error->message,
+                              XFCE_BUTTON_TYPE_MIXED, "window-close", _("_Close"), GTK_RESPONSE_ACCEPT,
+                              NULL);
+
+          g_clear_error (&error);
+        }
+
+      g_free(uri);
+      g_free(cmd);
+      g_object_unref(file);
+    }
+}
+
+
+static void
+garcon_gtk_menu_item_activate (GtkWidget      *mi,
+                               GarconMenuItem *item)
+{
+  GarconGtkMenu  *menu = g_object_get_data (G_OBJECT (mi), "GarconGtkMenu");
+  GdkEventButton *evt;
+  guint           button;
+  gboolean        right_click = FALSE;
+
+  evt = (GdkEventButton *)gtk_get_current_event();
+
+  /* See if we're trying to edit the launcher */
+   if(menu->priv->right_click_edits && evt && GDK_BUTTON_RELEASE == evt->type)
+    {
+      button = evt->button;
+
+      /* right click or Shift + left can optionally edit launchers */
+      if (button == 3 || (button == 1 && (evt->state & GDK_SHIFT_MASK)))
+        {
+          garcon_gtk_menu_item_edit_launcher (item);
+          right_click = TRUE;
+        }
+    }
+
+  if (!right_click)
+    {
+      /* normal action, launch the application */
+      garcon_gtk_menu_item_activate_real (mi, item, NULL);
+    }
+
+  if (evt)
+    {
+      gdk_event_free((GdkEvent*)evt);
+    }
+}
+
+
+
+static void
+garcon_gtk_menu_item_action_activate (GtkWidget            *mi,
+                                      GarconMenuItemAction *action)
+{
+  GarconMenuItem *item = g_object_get_data (G_OBJECT (action), "GarconMenuItem");
+
+  if (item == NULL)
+    {
+      g_critical ("garcon_gtk_menu_item_action_activate: Failed to get the GarconMenuItem\n");
+      return;
+    }
+
+  garcon_gtk_menu_item_activate_real (mi, item, action);
+}
+
+
+
 static void
 garcon_gtk_menu_item_drag_begin (GarconMenuItem *item,
                                  GdkDragContext *drag_context)
@@ -506,7 +652,7 @@ garcon_gtk_menu_load_icon (const gchar *icon_name)
 
   if (gtk_icon_theme_has_icon (icon_theme, icon_name))
     {
-      image = gtk_image_new_from_icon_name (icon_name, GTK_ICON_SIZE_MENU);
+	  pixbuf = gtk_icon_theme_load_icon (icon_theme, icon_name, size, 0, NULL);;
     }
   else
     {
@@ -542,18 +688,22 @@ garcon_gtk_menu_load_icon (const gchar *icon_name)
               g_free (name);
             }
         }
+    }
 
-      /* Turn the pixbuf into a gtk_image */
-      if (G_LIKELY (pixbuf))
-        {
-          /* scale the pixbuf down if it needs it */
-          GdkPixbuf *tmp = gdk_pixbuf_scale_simple (pixbuf, w, h, GDK_INTERP_BILINEAR);
-          g_object_unref (pixbuf);
-          pixbuf = tmp;
+  /* Turn the pixbuf into a gtk_image */
+  if (G_LIKELY (pixbuf))
+    {
+      /* scale the pixbuf down if it needs it */
+      GdkPixbuf *pixbuf_scaled = gdk_pixbuf_scale_simple (pixbuf, w, h, GDK_INTERP_BILINEAR);
+      g_object_unref (G_OBJECT (pixbuf));
 
-          image = gtk_image_new_from_pixbuf (pixbuf);
-          g_object_unref (G_OBJECT (pixbuf));
-        }
+      image = gtk_image_new_from_pixbuf (pixbuf_scaled);
+      g_object_unref (G_OBJECT (pixbuf_scaled));
+    }
+  else
+    {
+	  /* display the placeholder at least */
+	  image = gtk_image_new_from_icon_name (icon_name, GTK_ICON_SIZE_MENU);
     }
 
   return image;
@@ -561,13 +711,115 @@ garcon_gtk_menu_load_icon (const gchar *icon_name)
 
 
 
+static GtkWidget*
+garcon_gtk_menu_create_menu_item (GarconGtkMenu *menu,
+                                  const gchar *name,
+                                  const gchar *icon_name)
+{
+  GtkWidget *mi;
+  GtkWidget *box;
+  GtkWidget *image;
+  GtkWidget *label;
+
+  /* create item */
+  mi = gtk_menu_item_new ();
+  label = gtk_label_new (name);
+  gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
+#if GTK_CHECK_VERSION (3, 0, 0)
+  box = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
+  gtk_label_set_xalign (GTK_LABEL (label), 0.0);
+  gtk_widget_set_halign (label, GTK_ALIGN_START);
+#else
+  box = gtk_hbox_new (FALSE, 0);
+  gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5f);
+#endif
+
+  if (menu->priv->show_menu_icons)
+    {
+      image = garcon_gtk_menu_load_icon (icon_name);
+      gtk_widget_show (image);
+    }
+  else
+    {
+      image = gtk_image_new ();
+    }
+
+  /* Add the image and label to the box, add the box to the menu item */
+  gtk_box_pack_start (GTK_BOX (box), image, FALSE, FALSE, 0);
+  gtk_box_pack_start (GTK_BOX (box), label, TRUE, TRUE, 6);
+  gtk_widget_show_all (box);
+  gtk_container_add (GTK_CONTAINER (mi), box);
+
+  return mi;
+}
+
+
+
+static GtkWidget*
+garcon_gtk_menu_add_actions (GarconGtkMenu  *menu,
+                             GarconMenuItem *menu_item,
+                             GList          *actions,
+                             const gchar    *parent_icon_name)
+{
+  GtkWidget *submenu, *mi;
+  GList     *iter;
+
+  submenu = gtk_menu_new ();
+  gtk_menu_set_reserve_toggle_size (GTK_MENU (submenu), FALSE);
+
+  /* Add the parent item again, this time something the user can click to execute */
+  mi = garcon_gtk_menu_create_menu_item (menu, garcon_menu_item_get_name (menu_item), parent_icon_name);
+  gtk_menu_shell_append (GTK_MENU_SHELL (submenu), mi);
+  /* we need to store the GarconGtkMenu with this item so we can
+   * use it if the user wants to edit a menu item */
+  g_object_set_data (G_OBJECT (mi), "GarconGtkMenu", menu);
+  g_signal_connect (G_OBJECT (mi), "activate",
+                    G_CALLBACK (garcon_gtk_menu_item_activate), menu_item);
+  gtk_widget_show (mi);
+
+  /* Add all the individual actions to the menu */
+  for (iter = g_list_first(actions); iter != NULL; iter = g_list_next (iter))
+    {
+      GarconMenuItemAction *action = garcon_menu_item_get_action (menu_item, iter->data);
+      const gchar          *action_icon_name;
+
+      if (action == NULL)
+        continue;
+
+      /* If there's a custom icon associated with the action, use it.
+       * Otherwise default to the parent's icon.
+       */
+      action_icon_name = garcon_menu_item_action_get_icon_name (action);
+      if (action_icon_name == NULL)
+        {
+          action_icon_name = parent_icon_name;
+        }
+
+      mi = garcon_gtk_menu_create_menu_item (menu,
+                                             garcon_menu_item_action_get_name (action),
+                                             action_icon_name);
+
+      gtk_menu_shell_append (GTK_MENU_SHELL (submenu), mi);
+      g_signal_connect (G_OBJECT (mi), "activate",
+                        G_CALLBACK (garcon_gtk_menu_item_action_activate), action);
+      /* we need to store the parent associated with this item so we can
+       * activate it properly */
+      g_object_set_data (G_OBJECT (action), "GarconMenuItem", menu_item);
+      gtk_widget_show (mi);
+    }
+
+  return submenu;
+}
+
+
+
 static gboolean
 garcon_gtk_menu_add (GarconGtkMenu *menu,
                      GtkMenu       *gtk_menu,
                      GarconMenu    *garcon_menu)
 {
   GList               *elements, *li;
-  GtkWidget           *mi, *image;
+  GtkWidget           *mi;
   const gchar         *name, *icon_name;
   const gchar         *comment;
   GtkWidget           *submenu;
@@ -586,7 +838,7 @@ garcon_gtk_menu_add (GarconGtkMenu *menu,
 
       if (GARCON_IS_MENU_ITEM (li->data))
         {
-          GtkWidget *box, *label;
+          GList *actions = NULL;
 
           /* watch for changes */
           g_signal_connect_swapped (G_OBJECT (li->data), "changed",
@@ -606,40 +858,36 @@ garcon_gtk_menu_add (GarconGtkMenu *menu,
           if (G_UNLIKELY (name == NULL))
             continue;
 
-          /* create item */
-          mi = gtk_menu_item_new ();
-          label = gtk_label_new (name);
-          gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
-#if GTK_CHECK_VERSION (3, 0, 0)
-          box = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
-#else
-          box = gtk_hbox_new (FALSE, 0);
-          gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5f);
-#endif
+          icon_name = garcon_menu_item_get_icon_name (li->data);
+          if (STR_IS_EMPTY (icon_name))
+            icon_name = "applications-other";
 
-          if (menu->priv->show_menu_icons)
+          /* build the menu item */
+          mi = garcon_gtk_menu_create_menu_item (menu, name, icon_name);
+          gtk_menu_shell_append (GTK_MENU_SHELL (gtk_menu), mi);
+
+          /* if the menu item has actions such as "Private browsing mode"
+           * show them as well */
+          if (menu->priv->show_desktop_actions)
             {
-              icon_name = garcon_menu_item_get_icon_name (li->data);
-              if (STR_IS_EMPTY (icon_name))
-                icon_name = "applications-other";
+              actions = garcon_menu_item_get_actions (li->data);
+            }
 
-              image = garcon_gtk_menu_load_icon (icon_name);
-              gtk_widget_show (image);
+          if (actions != NULL)
+            {
+              submenu = garcon_gtk_menu_add_actions (menu, li->data, actions, icon_name);
+              gtk_menu_item_set_submenu (GTK_MENU_ITEM (mi), submenu);
+              g_list_free (actions);
             }
           else
             {
-              image = gtk_image_new ();
+              g_signal_connect (G_OBJECT (mi), "activate",
+                                G_CALLBACK (garcon_gtk_menu_item_activate), li->data);
+              /* we need to store the GarconGtkMenu with this item so we can
+               * use it if the user wants to edit a menu item */
+              g_object_set_data (G_OBJECT (mi), "GarconGtkMenu", menu);
             }
 
-          /* Add the image and label to the box, add the box to the menu item */
-          gtk_box_pack_start (GTK_BOX (box), image, FALSE, FALSE, 0);
-          gtk_box_pack_start (GTK_BOX (box), label, TRUE, TRUE, 6);
-          gtk_widget_show_all (box);
-          gtk_container_add (GTK_CONTAINER (mi), box);
-
-          gtk_menu_shell_append (GTK_MENU_SHELL (gtk_menu), mi);
-          g_signal_connect (G_OBJECT (mi), "activate",
-              G_CALLBACK (garcon_gtk_menu_item_activate), li->data);
           gtk_widget_show (mi);
 
           if (menu->priv->show_tooltips)
@@ -688,45 +936,24 @@ garcon_gtk_menu_add (GarconGtkMenu *menu,
           gtk_menu_set_reserve_toggle_size (GTK_MENU (submenu), FALSE);
           if (garcon_gtk_menu_add (menu, GTK_MENU (submenu), li->data))
             {
-              GtkWidget *box, *label;
-
               /* attach submenu */
               name = garcon_menu_element_get_name (li->data);
-              mi = gtk_menu_item_new ();
-              label = gtk_label_new (name);
-              gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);
-#if GTK_CHECK_VERSION (3, 0, 0)
-              box = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
-#else
-              box = gtk_hbox_new (FALSE, 0);
-              gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5f);
-#endif
 
-              if (menu->priv->show_menu_icons)
-                {
-                  icon_name = garcon_menu_element_get_icon_name (li->data);
-                  if (STR_IS_EMPTY (icon_name))
-                    icon_name = "applications-other";
+              icon_name = garcon_menu_element_get_icon_name (li->data);
+              if (STR_IS_EMPTY (icon_name))
+                icon_name = "applications-other";
 
-                  image = gtk_image_new_from_icon_name (icon_name, GTK_ICON_SIZE_MENU);
-                  gtk_widget_show (image);
-                }
-              else
-                {
-                  image = gtk_image_new ();
-                }
-
-              /* Add the image and label to the box, add the box to the menu item */
-              gtk_box_pack_start (GTK_BOX (box), image, FALSE, FALSE, 0);
-              gtk_box_pack_start (GTK_BOX (box), label, TRUE, TRUE, 6);
-              gtk_widget_show_all (box);
-              gtk_container_add (GTK_CONTAINER (mi), box);
+              /* build the menu item */
+              mi = garcon_gtk_menu_create_menu_item (menu, name, icon_name);
 
               gtk_menu_shell_append (GTK_MENU_SHELL (gtk_menu), mi);
               gtk_menu_item_set_submenu (GTK_MENU_ITEM (mi), submenu);
               g_signal_connect (G_OBJECT (submenu), "selection-done",
                   G_CALLBACK (garcon_gtk_menu_deactivate), menu);
               gtk_widget_show (mi);
+
+              /* atleast 1 visible child */
+              has_children = TRUE;
             }
           else
             {
@@ -747,7 +974,6 @@ static void
 garcon_gtk_menu_load (GarconGtkMenu *menu)
 {
   GError    *error = NULL;
-  GtkWidget *mi;
 
   g_return_if_fail (GARCON_GTK_IS_MENU (menu));
   g_return_if_fail (menu->priv->menu == NULL || GARCON_IS_MENU (menu->priv->menu));
@@ -757,13 +983,7 @@ garcon_gtk_menu_load (GarconGtkMenu *menu)
 
   if (garcon_menu_load (menu->priv->menu, NULL, &error))
     {
-      if (!garcon_gtk_menu_add (menu, GTK_MENU (menu), menu->priv->menu))
-        {
-          mi = gtk_menu_item_new_with_label (_("No applications found"));
-          gtk_menu_shell_append (GTK_MENU_SHELL (menu), mi);
-          gtk_widget_set_sensitive (mi, FALSE);
-          gtk_widget_show (mi);
-        }
+      garcon_gtk_menu_add (menu, GTK_MENU (menu), menu->priv->menu);
 
       /* watch for changes */
       g_signal_connect_swapped (G_OBJECT (menu->priv->menu), "reload-required",
@@ -807,15 +1027,10 @@ garcon_gtk_menu_new (GarconMenu *garcon_menu)
 
 
 /**
- * garcon_gtk_menu_get_menu:
+ * garcon_gtk_menu_set_menu:
  * @menu  : A #GarconGtkMenu
+ * @garcon_menu : The #GarconMenu to use
  *
- * The #GarconMenu used to create the #GtkMenu.
- *
- * The caller is responsible to releasing the returned #GarconMenu
- * using g_object_unref().
- *
- * Returns: the #GarconMenu for @menu.
  **/
 void
 garcon_gtk_menu_set_menu (GarconGtkMenu *menu,
@@ -834,7 +1049,7 @@ garcon_gtk_menu_set_menu (GarconGtkMenu *menu,
     }
 
   if (garcon_menu != NULL)
-    menu->priv->menu = g_object_ref (G_OBJECT (garcon_menu));
+    menu->priv->menu = GARCON_MENU (g_object_ref (G_OBJECT (garcon_menu)));
   else
     menu->priv->menu = NULL;
 
@@ -861,7 +1076,7 @@ garcon_gtk_menu_get_menu (GarconGtkMenu *menu)
 {
   g_return_val_if_fail (GARCON_GTK_IS_MENU (menu), NULL);
   if (menu->priv->menu != NULL)
-    return g_object_ref (G_OBJECT (menu->priv->menu));
+    return GARCON_MENU (g_object_ref (G_OBJECT (menu->priv->menu)));
   return NULL;
 }
 
@@ -979,3 +1194,77 @@ garcon_gtk_menu_get_show_tooltips (GarconGtkMenu *menu)
   g_return_val_if_fail (GARCON_GTK_IS_MENU (menu), FALSE);
   return menu->priv->show_tooltips;
 }
+
+/**
+ * garcon_gtk_menu_set_show_desktop_actions:
+ * @menu  : A #GarconGtkMenu
+ * @show_desktop_actions : Toggle showing the desktop actions in a submenu.
+ *
+ **/
+void
+garcon_gtk_menu_set_show_desktop_actions (GarconGtkMenu *menu,
+                                          gboolean       show_desktop_actions)
+{
+  g_return_if_fail (GARCON_GTK_IS_MENU (menu));
+
+  if (menu->priv->show_desktop_actions == show_desktop_actions)
+    return;
+
+  menu->priv->show_desktop_actions = !!show_desktop_actions;
+  g_object_notify_by_pspec (G_OBJECT (menu), menu_props[PROP_SHOW_DESKTOP_ACTIONS]);
+
+  garcon_gtk_menu_reload (menu);
+}
+
+
+
+/**
+ * garcon_gtk_menu_get_show_desktop_actions:
+ * @menu  : A #GarconGtkMenu
+ *
+ * Return value: if the desktop actions in a submenu
+ **/
+gboolean
+garcon_gtk_menu_get_show_desktop_actions (GarconGtkMenu *menu)
+{
+  g_return_val_if_fail (GARCON_GTK_IS_MENU (menu), FALSE);
+  return menu->priv->show_desktop_actions;
+}
+
+
+/**
+ * garcon_gtk_menu_set_right_click_edits:
+ * @menu  : A #GarconGtkMenu
+ * @enable_right_click_edits : Toggle showing wether to launch an editor
+ * when the menu is clicked with the secondary mouse button.
+ *
+ **/
+void
+garcon_gtk_menu_set_right_click_edits (GarconGtkMenu *menu,
+                                       gboolean       enable_right_click_edits)
+{
+  g_return_if_fail (GARCON_GTK_IS_MENU (menu));
+
+  if (menu->priv->right_click_edits == enable_right_click_edits)
+    return;
+
+  menu->priv->right_click_edits = !!enable_right_click_edits;
+  g_object_notify_by_pspec (G_OBJECT (menu), menu_props[PROP_RIGHT_CLICK_EDITS]);
+
+  garcon_gtk_menu_reload (menu);
+}
+
+
+
+/**
+ * garcon_gtk_menu_get_right_click_edits:
+ * @menu  : A #GarconGtkMenu
+ *
+ * Return value: if an editor will be launched on secondary mouse clicks.
+ **/
+gboolean
+garcon_gtk_menu_get_right_click_edits (GarconGtkMenu *menu)
+{
+  g_return_val_if_fail (GARCON_GTK_IS_MENU (menu), FALSE);
+  return menu->priv->right_click_edits;
+}
