@@ -3,6 +3,7 @@
  * Copyright (c) 2006-2010 Jannis Pohlmann <jannis@xfce.org>
  * Copyright (c) 2009-2010 Nick Schermer <nick@xfce.org>
  * Copyright (c) 2015      Danila Poyarkov <dannotemail@gmail.com>
+ * Copyright (c) 2017      Gregor Santner <gsantner@mailbox.org>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -81,8 +82,8 @@ static gboolean     garcon_menu_item_get_element_show_in_environment (GarconMenu
 static gboolean     garcon_menu_item_get_element_no_display          (GarconMenuElement      *element);
 static gboolean     garcon_menu_item_get_element_equal               (GarconMenuElement      *element,
                                                                       GarconMenuElement      *other);
-static gboolean     garcon_menu_item_category_lists_equal            (GList                  *categories1,
-                                                                      GList                  *categories2);
+static gboolean     garcon_menu_item_lists_equal                     (GList                  *list1,
+                                                                      GList                  *list2);
 
 
 
@@ -101,6 +102,9 @@ struct _GarconMenuItemPrivate
   /* List of categories */
   GList      *categories;
 
+  /* List of keywords */
+  GList      *keywords;
+
   /* Whether this application requires a terminal to be started in */
   guint       requires_terminal : 1;
 
@@ -137,8 +141,8 @@ struct _GarconMenuItemPrivate
   /* Working directory */
   gchar      *path;
 
-  /* Hash table for mapping action names to GarconMenuItemAction's */
-  GHashTable *actions;
+  /* List of application actions of type GarconMenuItemAction */
+  GList      *actions;
 
   /* Hidden value */
   guint       hidden : 1;
@@ -394,8 +398,6 @@ static void
 garcon_menu_item_init (GarconMenuItem *item)
 {
   item->priv = G_TYPE_INSTANCE_GET_PRIVATE (item, GARCON_TYPE_MENU_ITEM, GarconMenuItemPrivate);
-  item->priv->actions = g_hash_table_new_full (g_str_hash, g_str_equal, g_free,
-                                               (GDestroyNotify) garcon_menu_item_action_unref);
 }
 
 
@@ -418,12 +420,12 @@ garcon_menu_item_finalize (GObject *object)
   g_strfreev (item->priv->not_show_in);
 
   _garcon_g_list_free_full (item->priv->categories, g_free);
+  _garcon_g_list_free_full (item->priv->keywords, g_free);
+  _garcon_g_list_free_full (item->priv->actions, garcon_menu_item_action_unref);
 
   if (item->priv->file != NULL)
     g_object_unref (G_OBJECT (item->priv->file));
 
-  g_hash_table_unref (item->priv->actions);
-
   (*G_OBJECT_CLASS (garcon_menu_item_parent_class)->finalize) (object);
 }
 
@@ -670,18 +672,18 @@ garcon_menu_item_get_element_icon_name (GarconMenuElement *element)
 
 
 static gboolean
-garcon_menu_item_category_lists_equal (GList *categories1,
-                                       GList *categories2)
+garcon_menu_item_lists_equal (GList *list1,
+                              GList *list2)
 {
   gboolean  element_missing = FALSE;
   GList    *lp;
 
-  if (g_list_length (categories1) != g_list_length (categories2))
+  if (g_list_length (list1) != g_list_length (list2))
     return FALSE;
 
-  for (lp = categories1; !element_missing && lp != NULL; lp = lp->next)
+  for (lp = list1; !element_missing && lp != NULL; lp = lp->next)
     {
-      if (g_list_find_custom (categories2, lp->data, (GCompareFunc) g_strcmp0) == NULL)
+      if (g_list_find_custom (list2, lp->data, (GCompareFunc) g_strcmp0) == NULL)
         element_missing = TRUE;
     }
 
@@ -713,6 +715,7 @@ garcon_menu_item_new (GFile *file)
   GarconMenuItemAction *action = NULL;
   XfceRc               *rc;
   GList                *categories = NULL;
+  GList                *keywords = NULL;
   gchar                *filename;
   gboolean              terminal;
   gboolean              no_display;
@@ -801,6 +804,26 @@ garcon_menu_item_new (GFile *file)
           garcon_menu_item_set_categories (item, categories);
         }
 
+      /* Determine the keywords this application should be shown in */
+      str_list = xfce_rc_read_list_entry (rc, G_KEY_FILE_DESKTOP_KEY_KEYWORDS, ";");
+      if (G_LIKELY (str_list != NULL))
+        {
+          for (mt = str_list; *mt != NULL; ++mt)
+            {
+              /* Try to steal the values */
+              if (**mt != '\0')
+                keywords = g_list_prepend (keywords, *mt);
+              else
+                g_free (*mt);
+            }
+
+          /* Cleanup */
+          g_free (str_list);
+
+          /* Assign keywords list to the menu item */
+          garcon_menu_item_set_keywords (item, keywords);
+        }
+
       /* Set the rest of the private data directly */
       item->priv->only_show_in = xfce_rc_read_list_entry (rc, G_KEY_FILE_DESKTOP_KEY_ONLY_SHOW_IN, ";");
       item->priv->not_show_in = xfce_rc_read_list_entry (rc, G_KEY_FILE_DESKTOP_KEY_NOT_SHOW_IN, ";");
@@ -950,8 +973,10 @@ garcon_menu_item_reload_from_file (GarconMenuItem  *item,
   GarconMenuItemAction *action = NULL;
   gboolean              boolean;
   GList                *categories = NULL;
-  GList                *lp;
   GList                *old_categories = NULL;
+  GList                *keywords = NULL;
+  GList                *old_keywords = NULL;
+  GList                *lp;
   gchar               **mt;
   gchar               **str_list;
   const gchar          *string;
@@ -1002,7 +1027,7 @@ garcon_menu_item_reload_from_file (GarconMenuItem  *item,
     {
       if (G_LIKELY (item->priv->file != NULL))
         g_object_unref (G_OBJECT (item->priv->file));
-      item->priv->file = g_object_ref (G_OBJECT (file));
+      item->priv->file = G_FILE (g_object_ref (G_OBJECT (file)));
 
       g_object_notify (G_OBJECT (item), "file");
     }
@@ -1075,19 +1100,60 @@ garcon_menu_item_reload_from_file (GarconMenuItem  *item,
 
   if (affects_the_outside != NULL)
     {
-      if (!garcon_menu_item_category_lists_equal (old_categories, categories))
+      if (!garcon_menu_item_lists_equal (old_categories, categories))
         *affects_the_outside = TRUE;
 
       _garcon_g_list_free_full (old_categories, g_free);
     }
 
 
+  if (affects_the_outside != NULL)
+    {
+      /* create a deep copy the old keywords list */
+      old_keywords = g_list_copy (item->priv->keywords);
+      for (lp = old_keywords; lp != NULL; lp = lp->next)
+        lp->data = g_strdup (lp->data);
+    }
+
+  /* Determine the keywords this application should be shown in */
+  str_list = xfce_rc_read_list_entry (rc, G_KEY_FILE_DESKTOP_KEY_KEYWORDS, ";");
+  if (G_LIKELY (str_list != NULL))
+    {
+      for (mt = str_list; *mt != NULL; ++mt)
+        {
+          /* Try to steal the values */
+          if (**mt != '\0')
+            keywords = g_list_prepend (keywords, *mt);
+          else
+            g_free (*mt);
+        }
+
+      /* Cleanup */
+      g_free (str_list);
+
+      /* Assign keywords list to the menu item */
+      garcon_menu_item_set_keywords (item, keywords);
+    }
+  else
+    {
+      /* Assign empty keywords list to the menu item */
+      garcon_menu_item_set_keywords (item, NULL);
+    }
+
+  if (affects_the_outside != NULL)
+    {
+      if (!garcon_menu_item_lists_equal (old_keywords, keywords))
+        *affects_the_outside = TRUE;
+
+      _garcon_g_list_free_full (old_keywords, g_free);
+    }
+
   /* Set the rest of the private data directly */
   item->priv->only_show_in = xfce_rc_read_list_entry (rc, G_KEY_FILE_DESKTOP_KEY_ONLY_SHOW_IN, ";");
   item->priv->not_show_in = xfce_rc_read_list_entry (rc, G_KEY_FILE_DESKTOP_KEY_NOT_SHOW_IN, ";");
 
   /* Update application actions */
-  g_hash_table_remove_all (item->priv->actions);
+  _garcon_g_list_free_full (item->priv->actions, garcon_menu_item_action_unref);
   str_list = xfce_rc_read_list_entry (rc, G_KEY_FILE_DESKTOP_KEY_ACTIONS, ";");
   if (G_LIKELY (str_list != NULL))
     {
@@ -1261,6 +1327,34 @@ garcon_menu_item_set_categories (GarconMenuItem *item,
 
 
 
+GList*
+garcon_menu_item_get_keywords (GarconMenuItem *item)
+{
+  g_return_val_if_fail (GARCON_IS_MENU_ITEM (item), NULL);
+  return item->priv->keywords;
+}
+
+
+
+void
+garcon_menu_item_set_keywords (GarconMenuItem *item,
+                               GList          *keywords)
+{
+  g_return_if_fail (GARCON_IS_MENU_ITEM (item));
+
+  /* Abort if lists are equal */
+  if (G_UNLIKELY (item->priv->keywords == keywords))
+    return;
+
+  /* Free old list */
+  _garcon_g_list_free_full (item->priv->keywords, g_free);
+
+  /* Assign new list */
+  item->priv->keywords = keywords;
+}
+
+
+
 const gchar*
 garcon_menu_item_get_command (GarconMenuItem *item)
 {
@@ -1599,12 +1693,42 @@ garcon_menu_item_has_category (GarconMenuItem *item,
 
 
 
+gboolean
+garcon_menu_item_has_keyword (GarconMenuItem *item,
+                              const gchar    *keyword)
+{
+  GList   *iter;
+  gboolean found = FALSE;
+
+  g_return_val_if_fail (GARCON_IS_MENU_ITEM (item), FALSE);
+  g_return_val_if_fail (keyword != NULL, FALSE);
+
+  for (iter = item->priv->keywords; !found && iter != NULL; iter = g_list_next (iter))
+    if (g_strcmp0 (iter->data, keyword) == 0)
+      found = TRUE;
+
+  return found;
+}
+
+
+
 GList *
 garcon_menu_item_get_actions (GarconMenuItem *item)
 {
+  GList                *action_names = NULL;
+  GList                *iter;
+  GarconMenuItemAction *action;
+
   g_return_val_if_fail (GARCON_IS_MENU_ITEM (item), NULL);
-  
-  return g_hash_table_get_keys (item->priv->actions);
+
+  for (iter = item->priv->actions; iter != NULL ; iter = g_list_next (iter))
+    {
+      action = GARCON_MENU_ITEM_ACTION (iter->data);
+      action_names = g_list_prepend (action_names, (gchar*)garcon_menu_item_action_get_name (action));
+    }
+  action_names = g_list_reverse (action_names);
+
+  return action_names;
 }
 
 
@@ -1613,9 +1737,20 @@ GarconMenuItemAction *
 garcon_menu_item_get_action (GarconMenuItem *item,
                              const gchar    *action_name)
 {
+  GList                *iter;
+  GarconMenuItemAction *action;
+
   g_return_val_if_fail (GARCON_IS_MENU_ITEM (item), NULL);
-  
-  return g_hash_table_lookup (item->priv->actions, action_name);
+  g_return_val_if_fail (action_name != NULL, NULL);
+
+  for (iter = item->priv->actions; iter != NULL; iter = g_list_next (iter))
+    {
+      action = GARCON_MENU_ITEM_ACTION (iter->data);
+      if (g_strcmp0 (garcon_menu_item_action_get_name (action), action_name) == 0)
+        return (action);
+    }
+
+  return NULL;
 }
 
 
@@ -1626,14 +1761,39 @@ garcon_menu_item_set_action (GarconMenuItem       *item,
                              const gchar          *action_name,
                              GarconMenuItemAction *action)
 {
+  GList                *iter;
+  GarconMenuItemAction *old_action;
+  gboolean             found = FALSE;
+
   g_return_if_fail (GARCON_IS_MENU_ITEM (item));
   g_return_if_fail (GARCON_IS_MENU_ITEM_ACTION (action));
-  
-  /* Insert into the hash table and remove old action (if any) */
-  g_hash_table_replace (item->priv->actions, g_strdup (action_name), action);
 
-  /* Grab a reference on the action */
-  garcon_menu_item_action_ref (action);
+  /* If action name is found in list, then insert new action into the list and
+   * remove old action */
+  for (iter = item->priv->actions; !found && iter != NULL; iter = g_list_next (iter))
+    {
+      old_action = GARCON_MENU_ITEM_ACTION (iter->data);
+      if (g_strcmp0 (garcon_menu_item_action_get_name (old_action), action_name) == 0)
+        {
+           /* Release reference on action currently stored at action name */
+           garcon_menu_item_action_unref (old_action);
+
+           /* Replace action in list at action name and grab a reference */
+           iter->data = action;
+           garcon_menu_item_action_ref (action);
+
+           /* Set flag that action was found */
+           found = TRUE;
+        }
+    }
+
+  /* If action name was not found in list, then simply add it to list */
+  if (found == FALSE)
+    {
+      /* Add action to list and grab a reference */
+      item->priv->actions=g_list_append (item->priv->actions, action);
+      garcon_menu_item_action_ref (action);
+    }
 }
 
 
@@ -1642,9 +1802,21 @@ gboolean
 garcon_menu_item_has_action (GarconMenuItem  *item,
                              const gchar     *action_name)
 {
-  g_return_val_if_fail (GARCON_IS_MENU_ITEM (item), FALSE);
+  GList                *iter;
+  GarconMenuItemAction *action;
+  gboolean             found = FALSE;
 
-  return g_hash_table_contains (item->priv->actions, action_name);
+  g_return_val_if_fail (GARCON_IS_MENU_ITEM (item), FALSE);
+  g_return_val_if_fail (action_name != NULL, FALSE);
+
+  for (iter = item->priv->actions; !found && iter != NULL; iter = g_list_next (iter))
+    {
+      action = GARCON_MENU_ITEM_ACTION (iter->data);
+      if (g_strcmp0 (garcon_menu_item_action_get_name (action), action_name) == 0)
+        found = TRUE;
+    }
+
+  return found;
 }
 
 
@@ -1653,8 +1825,9 @@ gboolean
 garcon_menu_item_get_show_in_environment (GarconMenuItem *item)
 {
   const gchar *env;
-  guint        i;
+  guint        i, j;
   gboolean     show = TRUE;
+  gchar**      path = NULL;
 
   g_return_val_if_fail (GARCON_IS_MENU_ITEM (item), FALSE);
 
@@ -1666,20 +1839,29 @@ garcon_menu_item_get_show_in_environment (GarconMenuItem *item)
   if (G_UNLIKELY (env == NULL))
     return TRUE;
 
-  /* According to the spec there is either a OnlyShowIn or a NotShowIn list */
+  /* According to the spec there is either a OnlyShowIn or a NotShowIn list
+   * The environment can be multiple Desktop Names separated by a colons */
   if (G_UNLIKELY (item->priv->only_show_in != NULL))
     {
       /* Check if your environemnt is in OnlyShowIn list */
-      for (i = 0, show = FALSE; !show && item->priv->only_show_in[i] != NULL; i++)
-        if (g_strcmp0 (item->priv->only_show_in[i], env) == 0)
-          show = TRUE;
+      show = FALSE;
+      path = g_strsplit(env, ":", 0);
+      for (j = 0; path[j] != NULL; j++)
+        for (i = 0; !show && item->priv->only_show_in[i] != NULL; i++)
+          if (g_strcmp0 (item->priv->only_show_in[i], path[j]) == 0)
+            show = TRUE;
+      g_strfreev(path);
     }
   else if (G_UNLIKELY (item->priv->not_show_in != NULL))
     {
       /* Check if your environemnt is in NotShowIn list */
-      for (i = 0, show = TRUE; show && item->priv->not_show_in[i] != NULL; i++)
-        if (g_strcmp0 (item->priv->not_show_in[i], env) == 0)
-          show = FALSE;
+      show = TRUE;
+      path = g_strsplit(env, ":", 0);
+      for (j = 0; path[j] != NULL; j++)
+        for (i = 0; show && item->priv->not_show_in[i] != NULL; i++)
+          if (g_strcmp0 (item->priv->not_show_in[i], path[j]) == 0)
+            show = FALSE;
+      g_strfreev(path);
     }
 
   return show;
@@ -1691,8 +1873,9 @@ gboolean
 garcon_menu_item_only_show_in_environment (GarconMenuItem *item)
 {
   const gchar *env;
-  guint        i;
+  guint        i, j;
   gboolean     show = FALSE;
+  gchar**      path = NULL;
 
   g_return_val_if_fail (GARCON_IS_MENU_ITEM (item), FALSE);
 
@@ -1707,9 +1890,12 @@ garcon_menu_item_only_show_in_environment (GarconMenuItem *item)
   if (G_UNLIKELY (item->priv->only_show_in != NULL))
     {
       /* Check if your environemnt is in OnlyShowIn list */
-      for (i = 0, show = FALSE; !show && item->priv->only_show_in[i] != NULL; i++)
-        if (g_strcmp0 (item->priv->only_show_in[i], env) == 0)
-          show = TRUE;
+      show = FALSE;
+      path = g_strsplit(env, ":", 0);
+      for (j= 0; path[j] != NULL; j++)
+        for (i = 0; !show && item->priv->only_show_in[i] != NULL; i++)
+          if (g_strcmp0 (item->priv->only_show_in[i], path[j]) == 0)
+            show = TRUE;
     }
 
   return show;
@@ -1736,6 +1922,8 @@ garcon_menu_item_unref (GarconMenuItem *item)
 {
   g_return_if_fail (GARCON_IS_MENU_ITEM (item));
 
+  garcon_menu_item_decrement_allocated (item);
+
   /* Decrement the reference counter */
   g_object_unref (G_OBJECT (item));
 }
